from __future__ import annotations
from dataclasses import dataclass, field 
from typing import Any, Optional
from ._attachment import Attachment


@dataclass(eq=False)
class FileAttachment(Attachment):
    """Represents a file that has been sent as a Facebook attachment."""

    #: URL where you can download the file
    url: Optional[str] = field(default=None)
    #: Size of the file in bytes
    size: Any = field(default=None)
    #: Name of the file
    name: Optional[str] = field(default=None)
    #: Whether Facebook determines that this file may be harmful
    is_malicious: Optional[bool] = field(default=None)
    # Put here for backwards compatibility, so that the init argument order is preserved
    uid: Optional[int] = field(default=None)

    @classmethod
    def _from_graphql(cls, data)->FileAttachment:
        return cls(
            url=data.get("url"),
            name=data.get("filename"),
            is_malicious=data.get("is_malicious"),
            uid=data.get("message_file_fbid"),
        )


@dataclass(eq=False)
class AudioAttachment(Attachment):
    """Represents an audio file that has been sent as a Facebook attachment."""

    #: Name of the file
    filename: Optional[str] = field(default=None)
    #: URL of the audio file
    url: Optional[str] = field(default=None)
    #: Duration of the audio clip in milliseconds
    duration: Optional[int] = field(default=None)
    #: Audio type
    audio_type: Any = field(default=None)

    # Put here for backwards compatibility, so that the init argument order is preserved
    uid: Optional[int] = field(default=None)

    @classmethod
    def _from_graphql(cls, data)-> AudioAttachment:
        return cls(
            filename=data.get("filename"),
            url=data.get("playable_url"),
            duration=data.get("playable_duration_in_ms"),
            audio_type=data.get("audio_type"),
        )


@dataclass(eq=False, init=False)
class ImageAttachment(Attachment):
    """Represents an image that has been sent as a Facebook attachment.

    To retrieve the full image URL, use: `Client.fetchImageUrl`, and pass it the id of
    the image attachment.
    """

    #: The extension of the original image (e.g. ``png``)
    original_extension: Optional[str] = field(default=None)
    #: Width of original image
    width: Optional[int] = field(default=None)
    #: Height of original image
    height: Optional[int] = field(default=None)

    #: Whether the image is animated
    is_animated: Optional[bool] = field(default=None)

    #: URL to a thumbnail of the image
    thumbnail_url: Optional[str] = field(default=None)

    #: URL to a medium preview of the image
    preview_url: Optional[str] = field(default=None)
    #: Width of the medium preview image
    preview_width: Optional[int] = field(default=None)
    #: Height of the medium preview image
    preview_height: Optional[int] = field(default=None)

    #: URL to a large preview of the image
    large_preview_url: Optional[int] = field(default=None)
    #: Width of the large preview image
    large_preview_width: Optional[int] = field(default=None)
    #: Height of the large preview image
    large_preview_height: Optional[int] = field(default=None)

    #: URL to an animated preview of the image (e.g. for GIFs)
    animated_preview_url: Optional[str] = field(default=None)
    #: Width of the animated preview image
    animated_preview_width: Optional[int] = field(default=None)
    #: Height of the animated preview image
    animated_preview_height: Optional[int] = field(default=None)

    def __init__(
        self,
        original_extension=None,
        width=None,
        height=None,
        is_animated=None,
        thumbnail_url=None,
        preview=None,
        large_preview=None,
        animated_preview=None,
        **kwargs
    ):
        super(ImageAttachment, self).__init__(**kwargs)
        self.original_extension = original_extension
        if width is not None:
            width = int(width)
        self.width = width
        if height is not None:
            height = int(height)
        self.height = height
        self.is_animated = is_animated
        self.thumbnail_url = thumbnail_url

        if preview is None:
            preview = {}
        self.preview_url = preview.get("uri")
        self.preview_width = preview.get("width")
        self.preview_height = preview.get("height")

        if large_preview is None:
            large_preview = {}
        self.large_preview_url = large_preview.get("uri")
        self.large_preview_width = large_preview.get("width")
        self.large_preview_height = large_preview.get("height")

        if animated_preview is None:
            animated_preview = {}
        self.animated_preview_url = animated_preview.get("uri")
        self.animated_preview_width = animated_preview.get("width")
        self.animated_preview_height = animated_preview.get("height")

        def __post_init__(self):
            if self.width and not isinstance(self.width, int):
                self.width = int(self.width)
            if self.height and not isinstance(self.height, int):
                self.height = int(self.height)
            if self.preview_width and not isinstance(self.preview_width, int):
                self.preview_width = int(self.preview_width)
            if self.preview_height and not isinstance(self.preview_height, int):
                self.preview_height = int(self.preview_height)
            if self.large_preview_width and not isinstance(self.large_preview_width, int):
                self.large_preview_width = int(self.large_preview_width)
            if self.large_preview_height and not isinstance(self.large_preview_height, int):
                self.large_preview_height = int(self.large_preview_height)
            if self.animated_preview_width and not isinstance(self.animated_preview_width, int):
                self.animated_preview_width = int(self.animated_preview_width)
            if self.animated_preview_height and not isinstance(self.animated_preview_height, int):
                self.animated_preview_height = int(self.animated_preview_height)

    

    @classmethod
    def _from_graphql(cls, data)-> ImageAttachment:
        return cls(
            original_extension=data.get("original_extension")
            or (data["filename"].split("-")[0] if data.get("filename") else None),
            width=data.get("original_dimensions", {}).get("width"),
            height=data.get("original_dimensions", {}).get("height"),
            is_animated=data["__typename"] == "MessageAnimatedImage",
            thumbnail_url=data.get("thumbnail", {}).get("uri"),
            preview=data.get("preview") or data.get("preview_image"),
            large_preview=data.get("large_preview"),
            animated_preview=data.get("animated_image"),
            uid=data.get("legacy_attachment_id"),
        )

    @classmethod
    def _from_list(cls, data)-> ImageAttachment:
        data = data["node"]
        return cls(
            width=data["original_dimensions"].get("x"),
            height=data["original_dimensions"].get("y"),
            thumbnail_url=data["image"].get("uri"),
            large_preview=data["image2"],
            preview=data["image1"],
            uid=data["legacy_attachment_id"],
        )


@dataclass(eq=False, init=False)
class VideoAttachment(Attachment):
    """Represents a video that has been sent as a Facebook attachment."""

    #: Size of the original video in bytes
    size: Optional[int] = field(default=None)
    #: Width of original video
    width: Optional[int] = field(default=None)
    #: Height of original video
    height: Optional[int] = field(default=None)
    #: Length of video in milliseconds
    duration: Optional[int] = field(default=None)
    #: URL to very compressed preview video
    preview_url: Optional[int] = field(default=None)
    #: URL to a small preview image of the video
    small_image_url: Optional[str] = field(default=None)
    #: Width of the small preview image
    small_image_width: Optional[int] = field(default=None)
    #: Height of the small preview image
    small_image_height: Optional[int] = field(default=None)
    #: URL to a medium preview image of the video
    medium_image_url: Optional[str] = field(default=None)
    #: Width of the medium preview image
    medium_image_width: Optional[int] = field(default=None)
    #: Height of the medium preview image
    medium_image_height: Optional[int] = field(default=None)

    #: URL to a large preview image of the video
    large_image_url: Optional[str] = field(default=None)
    #: Width of the large preview image
    large_image_width: Optional[int] = field(default=None)
    #: Height of the large preview image
    large_image_height: Optional[int] = field(default=None)

    def __init__(
        self,
        size=None,
        width=None,
        height=None,
        duration=None,
        preview_url=None,
        small_image=None,
        medium_image=None,
        large_image=None,
        **kwargs
    ):
        super(VideoAttachment, self).__init__(**kwargs)
        self.size = size
        self.width = width
        self.height = height
        self.duration = duration
        self.preview_url = preview_url

        if small_image is None:
            small_image = {}
        self.small_image_url = small_image.get("uri")
        self.small_image_width = small_image.get("width")
        self.small_image_height = small_image.get("height")

        if medium_image is None:
            medium_image = {}
        self.medium_image_url = medium_image.get("uri")
        self.medium_image_width = medium_image.get("width")
        self.medium_image_height = medium_image.get("height")

        if large_image is None:
            large_image = {}
        self.large_image_url = large_image.get("uri")
        self.large_image_width = large_image.get("width")
        self.large_image_height = large_image.get("height")

    @classmethod
    def _from_graphql(cls, data)-> VideoAttachment:
        return cls(
            width=data.get("original_dimensions", {}).get("width"),
            height=data.get("original_dimensions", {}).get("height"),
            duration=data.get("playable_duration_in_ms"),
            preview_url=data.get("playable_url"),
            small_image=data.get("chat_image"),
            medium_image=data.get("inbox_image"),
            large_image=data.get("large_image"),
            uid=data.get("legacy_attachment_id"),
        )

    @classmethod
    def _from_subattachment(cls, data)-> VideoAttachment:
        media = data["media"]
        return cls(
            duration=media.get("playable_duration_in_ms"),
            preview_url=media.get("playable_url"),
            medium_image=media.get("image"),
            uid=data["target"].get("video_id"),
        )

    @classmethod
    def _from_list(cls, data)-> VideoAttachment:
        data = data["node"]
        return cls(
            width=data["original_dimensions"].get("x"),
            height=data["original_dimensions"].get("y"),
            small_image=data["image"],
            medium_image=data["image1"],
            large_image=data["image2"],
            uid=data["legacy_attachment_id"],
        )


def graphql_to_attachment(data):
    _type = data["__typename"]
    if _type in ["MessageImage", "MessageAnimatedImage"]:
        return ImageAttachment._from_graphql(data)
    elif _type == "MessageVideo":
        return VideoAttachment._from_graphql(data)
    elif _type == "MessageAudio":
        return AudioAttachment._from_graphql(data)
    elif _type == "MessageFile":
        return FileAttachment._from_graphql(data)

    return Attachment(uid=data.get("legacy_attachment_id"))


def graphql_to_subattachment(data):
    target = data.get("target")
    type_ = target.get("__typename") if target else None

    if type_ == "Video":
        return VideoAttachment._from_subattachment(data)

    return None



from __future__ import annotations

from ._attachment import Attachment
from dataclasses import dataclass
from typing import Optional

@dataclass(eq=False, init=False)
class Sticker(Attachment):
    """
    Represents a Facebook sticker that has been sent to a thread as an attachment.
    """

    #: The sticker-pack's ID
    pack: Optional[str] = None
    #: Whether the sticker is animated
    is_animated: bool = False

    # If the sticker is animated, the following should be present
    #: URL to a medium spritemap
    medium_sprite_image: Optional[str] = None
    #: URL to a large spritemap
    large_sprite_image: Optional[str] = None
    #: The amount of frames present in the spritemap pr. row
    frames_per_row: Optional[int] = None
    #: The amount of frames present in the spritemap pr. column
    frames_per_col: Optional[int] = None
    #: The frame rate the spritemap is intended to be played in
    frame_rate: Optional[float] = None

    #: URL to the sticker's image
    url: Optional[str] = None
    #: Width of the sticker
    width: Optional[int] = None
    #: Height of the sticker
    height: Optional[int] = None
    #: The sticker's label/name
    label: Optional[str] = None

    def __init__(self, uid: Optional[int] = None):
        super(Sticker, self).__init__(uid=uid)

    @classmethod
    def _from_graphql(cls, data)-> Sticker:
        if not data:
            return None
        self = cls(uid=data["id"])
        if data.get("pack"):
            self.pack = data["pack"].get("id")
        if data.get("sprite_image"):
            self.is_animated = True
            self.medium_sprite_image = data["sprite_image"].get("uri")
            self.large_sprite_image = data["sprite_image_2x"].get("uri")
            self.frames_per_row = data.get("frames_per_row")
            self.frames_per_col = data.get("frames_per_column")
            self.frame_rate = data.get("frame_rate")
        self.url = data.get("url")
        self.width = data.get("width")
        self.height = data.get("height")
        if data.get("label"):
            self.label = data["label"]
        return self



import json
from typing import List, Union, Dict, Any
from . import _util
from .models import FBchatException

# Lord Yuuta doesnt copy he invents

def split_json_objects(json_string: str)-> List[Dict[str, Any]]:
    # Remove very long spaces
    json_string = "".join(json_string.split())
    # Json string contains list of json objects without the `[` and `]` 

    # We need to parse them into a list of Dict in order to use json.loads()
    depth = 0
    start = 0
    results = []
    
    for i, char in enumerate(json_string):
        if char == '{':
            if depth == 0:
                start = i
            depth += 1
        elif char == '}':
            depth -= 1
            if depth == 0:
                results.append(json_string[start:i+1])
    
    return [json.loads(chunk) for chunk in results]



def queries_to_json(*queries)-> str:
    """
    queries should be a list/tuple of GraphQL objects
    returns Json string.
    """
    rtn = {
        f"q{i}": query for i, query in enumerate(queries)
    }
    return json.dumps(rtn)


def response_to_json(content)-> List[Union[None, Dict[str, Any]]]:
    """
    Process JSON-like content and extract data into a structured list.

    Args:
        content (str): The raw JSON string to process.

    Returns:
        List[Union[None, Dict[str, Any]]]: A list where each index corresponds to a parsed JSON object.
        - None: If no valid data is found for that index.
        - Dict[str, Any]: If valid JSON data is found at that index.
    """
    # Remove unwanted characters from the content
    content = _util.strip_json_cruft(content)
    # Parse JSON objects
    try:
        parsed_objects = split_json_objects(content)
    except Exception as e:
        raise FBchatException(f"Error while parsing JSON: {repr(content)}") from e
    # Initialize the result list
    results = [None] * len(parsed_objects)
    for obj in parsed_objects:
        # Skip objects with "error_results"
        if "error_results" in obj:
            results.pop()  # Remove the last element
            continue
        # Handle potential errors in the payload and GraphQL
        _util.handle_payload_error(obj)
        [(key, value)] = obj.items()
        _util.handle_graphql_errors(value)
        # Extract response or data
        index = int(key[1:])  # Extract numeric part from key (e.g., "q0" -> 0)
        if "response" in value:
            results[index] = value["response"]
        else:
            results[index] = value["data"]
    return results  #type: ignore


def from_query(query, params):
    return {"priority": 0, "q": query, "query_params": params}


def from_query_id(query_id, params):
    return {"query_id": query_id, "query_params": params}


def from_doc(doc, params):
    return {"doc": doc, "query_params": params}


def from_doc_id(doc_id, params):
    return {"doc_id": doc_id, "query_params": params}


FRAGMENT_USER = """
QueryFragment User: User {
    id,
    name,
    first_name,
    last_name,
    profile_picture.width(<pic_size>).height(<pic_size>) {
        uri
    },
    is_viewer_friend,
    url,
    gender,
    viewer_affinity
}
"""

FRAGMENT_GROUP = """
QueryFragment Group: MessageThread {
    name,
    thread_key {
        thread_fbid
    },
    image {
        uri
    },
    is_group_thread,
    all_participants {
        nodes {
            messaging_actor {
                id
            }
        }
    },
    customization_info {
        participant_customizations {
            participant_id,
            nickname
        },
        outgoing_bubble_color,
        emoji
    },
    thread_admins {
        id
    },
    group_approval_queue {
        nodes {
            requester {
                id
            }
        }
    },
    approval_mode,
    joinable_mode {
        mode,
        link
    },
    event_reminders {
        nodes {
            id,
            lightweight_event_creator {
                id
            },
            time,
            location_name,
            event_title,
            event_reminder_members {
                edges {
                    node {
                        id
                    },
                    guest_list_state
                }
            }
        }
    }
}
"""

FRAGMENT_PAGE = """
QueryFragment Page: Page {
    id,
    name,
    profile_picture.width(32).height(32) {
        uri
    },
    url,
    category_type,
    city {
        name
    }
}
"""

SEARCH_USER = (
    """
Query SearchUser(<search> = '', <limit> = 10) {
    entities_named(<search>) {
        search_results.of_type(user).first(<limit>) as users {
            nodes {
                @User
            }
        }
    }
}
"""
    + FRAGMENT_USER
)

SEARCH_GROUP = (
    """
Query SearchGroup(<search> = '', <limit> = 10, <pic_size> = 32) {
    viewer() {
        message_threads.with_thread_name(<search>).last(<limit>) as groups {
            nodes {
                @Group
            }
        }
    }
}
"""
    + FRAGMENT_GROUP
)

SEARCH_PAGE = (
    """
Query SearchPage(<search> = '', <limit> = 10) {
    entities_named(<search>) {
        search_results.of_type(page).first(<limit>) as pages {
            nodes {
                @Page
            }
        }
    }
}
"""
    + FRAGMENT_PAGE
)

SEARCH_THREAD = (
    """
Query SearchThread(<search> = '', <limit> = 10) {
    entities_named(<search>) {
        search_results.first(<limit>) as threads {
            nodes {
                __typename,
                @User,
                @Group,
                @Page
            }
        }
    }
}
"""
    + FRAGMENT_USER
    + FRAGMENT_GROUP
    + FRAGMENT_PAGE
)



from __future__ import annotations

import aiohttp
import re
import random

from yarl import URL
from typing import Any, Dict

from dataclasses import dataclass, field
from .n_util import * 

from . import _graphql 
from .models import _exception

__all__ = ["State"]

def get_session(cookies: Dict | None = None)-> aiohttp.ClientSession:
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Referer": "https://www.facebook.com/",
        "Host": "www.facebook.com",
        "Origin": "https://www.facebook.com",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36",
        "Connection": "keep-alive",
        "Sec-Fetch-Site": "same-origin",
        "Sec-Fetch-User": "?1",
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
        "Accept-Encoding": "gzip", #, deflate, br",
        "Accept-Language": "en-US,en;q=0.9"
    } 
    session = aiohttp.ClientSession(headers=headers, cookies=cookies)
    #    save_cookies_to_session(session, cookies)

    return session


def get_user_id(session: aiohttp.ClientSession)-> str:
        cookies = session.cookie_jar.filter_cookies(URL("https://www.facebook.com"))
        CookiesDict = {name: cookie.value for name, cookie in cookies.items()}
        return str(CookiesDict.get("c_user"))


def client_id_factory()-> str:
      return hex(int(random.random() * 2 ** 31))[2:]


@dataclass
class State:

    HEADERS = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Referer": "https://www.facebook.com/",
        "Host": "www.facebook.com",
        "Origin": "https://www.facebook.com",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36",
        "Connection": "keep-alive",
        "Sec-Fetch-Site": "same-origin",
        "Sec-Fetch-User": "?1",
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
        "Accept-Encoding": "gzip", #, deflate, br",
        "Accept-Language": "en-US,en;q=0.9"
    }
    user_id: str = field()
    _fb_dtsg: str = field()
    _revision: int | None = field()
    _client_id: str = field(default_factory=client_id_factory)
    _session: aiohttp.client.ClientSession = field(default_factory=get_session)
    _logout_h: Any = field(default=None)
    _counter: int = field(default=0)

    def get_params(self):
        self._counter += 1

        return {
            "__a": 1,   
            "__req": decimal_to_base36(self._counter),
            "__rev": self._revision,
            "fb_dtsg": self._fb_dtsg
        }

    @classmethod
    async def from_cookies(cls, cookies)-> State:
        session = get_session(cookies=cookies)
        return await cls.from_session(session)


    @classmethod
    async def from_json_cookies(cls, json_cookies_path: str)-> State:
        """pass the saved facebook cookies json file to

        Args:
            json_cookies_path (str): the saved facebook cookies json file.

        Returns:
            State: returns an instance of State class with session.
        """
        cookies = get_cookie(json_cookies_path)
        session = get_session(cookies=cookies)
        return await cls.from_session(session)

    @classmethod
    async def from_session(cls, session: aiohttp.client.ClientSession):
        user_id = get_user_id(session)
        response = await session.get("https://www.facebook.com/")
        results = await response.text()

        fb_dtsg = re.search(r'"DTSGInitialData".*?"token":"(.*?)"', results)
        if fb_dtsg:
            fb_dtsg = fb_dtsg.group(1)
        else:
            raise ValueError("fb_dtsg not found.")
        clientRevision = re.search(r'client_revision":(\d+)', results)
        if clientRevision:
            clientRevision = int(clientRevision.group(1))

        return cls(
            user_id = user_id,
            _fb_dtsg = fb_dtsg,
            _revision = clientRevision,
            _session = session,
            _logout_h = None
        )

    async def is_logged_in(self)-> bool:
        """Sends request to facebook to check login status"""
        try:
            req = await self._session.get(prefix_url("/profile.php"), allow_redirects=False)
        except aiohttp.ClientError as e:
            raise Exception("Client Error: ", e)
        location = req.headers.get("Location")
        return str(location) == f"https://www.facebook.com/profile.php?id={self.user_id}"



    def get_cookies(self):
        cookies = self._session.cookie_jar.filter_cookies(URL("https://www.facebook.com"))
        return {name: cookie.value for name, cookie in cookies.items()}


    
    async def _do_refresh(self):
        # TODO: Raise the error instead, and make the user do the refresh manually
    # It may be a bad idea to do this in an exception handler, if you have a better method
        new = await State.from_session(session=self._session)
        self.user_id = new.user_id
        self._fb_dtsg = new._fb_dtsg
        self._revision = new._revision
        self._counter = new._counter
        self._logout_h = new._logout_h

    async def _get(self, url: str, params: Dict, error_retries=3):
        params.update(self.get_params())
        response = await self._session.get(URL(prefix_url(url)), params=params)
        content = await check_request(response)
        json_data = to_json(content)
        try:
            handle_payload_error(json_data)
        except:
            if error_retries > 0:
                await self._do_refresh()
                return await self._get(url, params, error_retries=error_retries)
            raise
        return json_data


    async def _post(self, url: str, data, files=None, as_graphql=False, error_retries=3)-> dict:
        data.update(self.get_params())
        headers = self.HEADERS
        self._session.headers.clear()
        if files:
            formData = aiohttp.FormData()
            for key, value in data.items():
                formData.add_field(key, str(value))
            for key, (name, file, content_type) in files.items():
                formData.add_field(key, file, filename=name, content_type=content_type)

            data = formData 
            # Let aiohttp automatically handle Content-Type
            headers.pop("Content-Type", None)

        response = await self._session.post(URL(prefix_url(url)), data=data, headers=headers)

        content = await check_request(response)
        self._session._default_headers = self._session._prepare_headers(self.HEADERS)
        try:
            if as_graphql:
                return _graphql.response_to_json(content)
            else:
                j = to_json(content)
                handle_payload_error(j)
                return j
        except aiohttp.ClientResponseError as e:
            if error_retries > 0:
                await self._do_refresh()
                return await self._post(url, data=data, files=files, as_graphql=as_graphql, error_retries=(error_retries - 1))
            raise e

    async def _payload_post(self, url, data, files=None):
        response: Dict = await self._post(url, data=data, files=files)
        try:
            return response["payload"]
            
        except (KeyError, TypeError):
            raise aiohttp.ClientPayloadError(f"missing payload: {response}")
            
    async def _graphql_requests(self, *queries):
        data = {
            "method": "GET",
            "response_format": "json",
            "queries": _graphql.queries_to_json(*queries),
        }
        return await self._post("/api/graphqlbatch/", data, as_graphql=True)

    async def _upload(self, files, voice_clip=False):

        file_dict = {f"upload_{i}": file for i, file in enumerate(files)}
        data = {"voice_clip": voice_clip}
        json_response: dict = await self._payload_post("https://upload.facebook.com/ajax/mercury/upload.php", data, files=file_dict)
        if len(json_response["metadata"]) != len(files):
            return f"Some files could not be uploaded: {json_response}, {files}"
        to_return = [
            (data[1][mimetype_to_key(data[1]["filetype"])], data[1]["filetype"]) 
            for data in json_response["metadata"].items()
        ]
        return to_return

    async def do_send_request(self, data):
        offline_threading_id = generateOfflineThreadingID()
        data["client"] = "mercury"
        data["author"] = f"fbid:{self.user_id}"
        data["timestamp"] = now_time()
        data["source"] = "source:chat:web"
        data["offline_threading_id"] = offline_threading_id
        data["message_id"] = offline_threading_id
        data["threading_id"] = generateMessageID(self._client_id)
        data["ephemeral_ttl_mode:"] = "0"
        json_response = await self._post("/messaging/send/", data)        
        # update JS token if received in response
        fb_dtsg = get_jsmods_require(json_response, 2)
        if fb_dtsg is not None:
            self._fb_dtsg = fb_dtsg

        try:
            message_ids = [
                (action["message_id"], action["thread_fbid"])
                for action in json_response["payload"]["actions"]
                if "message_id" in action
            ]
            if len(message_ids) != 1:
                print("Got multiple message ids' back: {}".format(message_ids))
            return message_ids[0]
        except (KeyError, IndexError, TypeError):
            raise _exception.FBchatException(
                "Error when sending message: "
                f"No message IDs could be found: {json_response}"
            )




from __future__ import annotations

from typing import Dict, Optional
from dataclasses import dataclass, field

from ._plan import Plan
from ._core import CustomEnum


class ThreadType(CustomEnum):
    """Used to specify what type of Facebook thread is being used.

    See :ref:`intro_threads` for more info.
    """

    USER = 1
    GROUP = 2
    ROOM = 2
    PAGE = 3

    def _to_class(self):
        """Convert this enum value to the corresponding class."""
        from . import _user, _group, _page

        return {
            ThreadType.USER: _user.User,
            ThreadType.GROUP: _group.Group,
            ThreadType.ROOM: _group.Room,
            ThreadType.PAGE: _page.Page,
        }[self]


class ThreadLocation(CustomEnum):
    """Used to specify where a thread is located (inbox, pending, archived, other)."""

    INBOX = "INBOX"
    PENDING = "PENDING"
    ARCHIVED = "ARCHIVED"
    OTHER = "OTHER"


class ThreadColor(CustomEnum):
    """Used to specify a thread colors."""

    MESSENGER_BLUE = "#0084ff"
    VIKING = "#44bec7"
    GOLDEN_POPPY = "#ffc300"
    RADICAL_RED = "#fa3c4c"
    SHOCKING = "#d696bb"
    PICTON_BLUE = "#6699cc"
    FREE_SPEECH_GREEN = "#13cf13"
    PUMPKIN = "#ff7e29"
    LIGHT_CORAL = "#e68585"
    MEDIUM_SLATE_BLUE = "#7646ff"
    DEEP_SKY_BLUE = "#20cef5"
    FERN = "#67b868"
    CAMEO = "#d4a88c"
    BRILLIANT_ROSE = "#ff5ca1"
    BILOBA_FLOWER = "#a695c7"
    TICKLE_ME_PINK = "#ff7ca8"
    MALACHITE = "#1adb5b"
    RUBY = "#f01d6a"
    DARK_TANGERINE = "#ff9c19"
    BRIGHT_TURQUOISE = "#0edcde"

    @classmethod
    def _from_graphql(cls, color)-> Optional[ThreadColor]:
        if color is None:
            return None
        if not color:
            return cls.MESSENGER_BLUE
        color = color[2:]  # Strip the alpha value
        value = "#{}".format(color.lower())
        return cls._extend_if_invalid(value)


@dataclass(eq=False, init=False)
class Thread:
    """Represents a Facebook thread."""

    #: The unique identifier of the thread. Can be used a ``thread_id``. See :ref:`intro_threads` for more info
    uid: str = field()
    #: Specifies the type of thread. Can be used a ``thread_type``. See :ref:`intro_threads` for more info
    type: Optional[ThreadType] = None
    #: A URL to the thread's picture
    photo: Optional[str] = None
    #: The name of the thread
    name: Optional[str] = None
    #: Timestamp of last message
    last_message_timestamp: Optional[int] = None
    #: Number of messages in the thread
    message_count: Optional[int] = None
    #: Set :class:`Plan`
    plan: Optional[Plan] = None
    

    def __post_init__(self):
        self.uid = str(self.uid)

    @staticmethod
    def _parse_customization_info(data)-> Dict:
        if data is None or data.get("customization_info") is None:
            return {}
        info = data["customization_info"]

        rtn = {
            "emoji": info.get("emoji"),
            "color": ThreadColor._from_graphql(info.get("outgoing_bubble_color")),
        }
        if (
            data.get("thread_type") == "GROUP"
            or data.get("is_group_thread")
            or data.get("thread_key", {}).get("thread_fbid")
        ):
            rtn["nicknames"] = {}
            for k in info.get("participant_customizations", []):
                rtn["nicknames"][k["participant_id"]] = k.get("nickname")
        elif info.get("participant_customizations"):
            uid = data.get("thread_key", {}).get("other_user_id") or data.get("id")
            pc = info["participant_customizations"]
            if len(pc) > 0:
                if pc[0].get("participant_id") == uid:
                    rtn["nickname"] = pc[0].get("nickname")
                else:
                    rtn["own_nickname"] = pc[0].get("nickname")
            if len(pc) > 1:
                if pc[1].get("participant_id") == uid:
                    rtn["nickname"] = pc[1].get("nickname")
                else:
                    rtn["own_nickname"] = pc[1].get("nickname")
        return rtn

    def _to_send_data(self)-> Dict[str, str]:
        # TODO: Only implement this in subclasses
        return {"other_user_fbid": self.uid}



from __future__ import annotations
import json
from typing import Dict, Optional
from ._core import CustomEnum
from dataclasses import dataclass, field

class GuestStatus(CustomEnum):
    INVITED = 1
    GOING = 2
    DECLINED = 3


@dataclass(eq=False)
class Plan:
    """Represents a plan."""

    #: ID of the plan
    uid: Optional[str] = field(default=None, init=False)
    #: Plan time (timestamp), only precise down to the minute
    time: Optional[int]
    #: Plan title
    title: str
    #: Plan location name
    location: Optional[str] = None
    #: Plan location ID
    location_id: Optional[str] = None
    #: ID of the plan creator
    author_id: Optional[str] = field(default=None, init=False)
    #: Dictionary of `User` IDs mapped to their `GuestStatus`
    guests: Optional[Dict[str, GuestStatus]] = field(default=None, init=False)

    def __post_init__(self):
        if self.time:
            self.time = int(self.time)
        self.location = self.location or ""
        self.location_id = self.location_id or ""

    @property
    def going(self):
        """List of the `User` IDs who will take part in the plan."""
        return [
            id_
            for id_, status in (self.guests or {}).items()
            if status is GuestStatus.GOING
        ]

    @property
    def declined(self):
        """List of the `User` IDs who won't take part in the plan."""
        return [
            id_
            for id_, status in (self.guests or {}).items()
            if status is GuestStatus.DECLINED
        ]

    @property
    def invited(self):
        """List of the `User` IDs who are invited to the plan."""
        return [
            id_
            for id_, status in (self.guests or {}).items()
            if status is GuestStatus.INVITED
        ]

    @classmethod
    def _from_pull(cls, data)->Plan:
        rtn = cls(
            time=data.get("event_time"),
            title=data.get("event_title"),
            location=data.get("event_location_name"),
            location_id=data.get("event_location_id"),
        )
        rtn.uid = data.get("event_id")
        rtn.author_id = data.get("event_creator_id")
        rtn.guests = {
            x["node"]["id"]: GuestStatus[x["guest_list_state"]]
            for x in json.loads(data["guest_state_list"])
        }
        return rtn

    @classmethod
    def _from_fetch(cls, data)-> Plan:
        rtn = cls(
            time=data.get("event_time"),
            title=data.get("title"),
            location=data.get("location_name"),
            location_id=str(data["location_id"]) if data.get("location_id") else None,
        )
        rtn.uid = data.get("oid")
        rtn.author_id = data.get("creator_id")
        rtn.guests = {id_: GuestStatus[s] for id_, s in data["event_members"].items()}
        return rtn

    @classmethod
    def _from_graphql(cls, data)-> Plan:
        rtn = cls(
            time=data.get("time"),
            title=data.get("event_title"),
            location=data.get("location_name"),
        )
        rtn.uid = data.get("id")
        rtn.author_id = data["lightweight_event_creator"].get("id")
        rtn.guests = {
            x["node"]["id"]: GuestStatus[x["guest_list_state"]]
            for x in data["event_reminder_members"]["edges"]
        }
        return rtn



from dataclasses import dataclass, field
from typing import Any, Dict, Optional

@dataclass(eq=False)
class QuickReply:
    """Represents a quick reply."""

    #: Payload of the quick reply
    payload: Optional[Dict] = field(default=None)
    #: External payload for responses
    external_payload: Optional[Dict] = field(default=None, init=False)
    #: Additional data
    data: Any = field(default=None)
    #: Whether it's a response for a quick reply
    is_response: bool = field(default=False)


@dataclass(eq=False, init=False)
class QuickReplyText(QuickReply):
    """Represents a text quick reply."""

    #: Title of the quick reply
    title: Optional[str] = field(default=None)
    #: URL of the quick reply image (optional)
    image_url: Optional[str] = field(default=None)
    #: Type of the quick reply
    _type = "text"

    def __init__(self, title=None, image_url=None, **kwargs):
        super(QuickReplyText, self).__init__(**kwargs)
        self.title = title
        self.image_url = image_url


@dataclass(eq=False, init=False)
class QuickReplyLocation(QuickReply):
    """Represents a location quick reply (Doesn't work on mobile)."""

    #: Type of the quick reply
    _type = "location"

    def __init__(self, **kwargs):
        super(QuickReplyLocation, self).__init__(**kwargs)
        self.is_response = False


@dataclass(eq=False, init=False)
class QuickReplyPhoneNumber(QuickReply):
    """Represents a phone number quick reply (Doesn't work on mobile)."""

    #: URL of the quick reply image (optional)
    image_url: Optional[str] = field(default=None)
    #: Type of the quick reply
    _type = "user_phone_number"

    def __init__(self, image_url=None, **kwargs):
        super(QuickReplyPhoneNumber, self).__init__(**kwargs)
        self.image_url = image_url


@dataclass(eq=False, init=False)
class QuickReplyEmail(QuickReply):
    """Represents an email quick reply (Doesn't work on mobile)."""

    #: URL of the quick reply image (optional)
    image_url: Optional[str] = field(default=None)
    #: Type of the quick reply
    _type = "user_email"

    def __init__(self, image_url=None, **kwargs):
        super(QuickReplyEmail, self).__init__(**kwargs)
        self.image_url = image_url


def graphql_to_quick_reply(q, is_response: bool = False):
    data = dict()
    _type = q.get("content_type").lower()
    if q.get("payload"):
        data["payload"] = q["payload"]
    if q.get("data"):
        data["data"] = q["data"]
    if q.get("image_url") and _type is not QuickReplyLocation._type:
        data["image_url"] = q["image_url"]
    data["is_response"] = is_response

    if _type == QuickReplyText._type:
        if q.get("title") is not None:
            data["title"] = q["title"]
        return QuickReplyText(**data)
    elif _type == QuickReplyLocation._type:
        return QuickReplyLocation(**data)
    elif _type == QuickReplyPhoneNumber._type:
        return QuickReplyPhoneNumber(**data)
    elif _type == QuickReplyEmail._type:
        return QuickReplyEmail(**data)
    



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Optional
from ._attachment import Attachment
from .. import _util


@dataclass(eq=False)
class LocationAttachment(Attachment):
    """Represents a user location.

    Latitude and longitude OR address is provided by Facebook.
    """

    #: Latitude of the location
    latitude: Optional[float] = None
    #: Longitude of the location
    longitude: Optional[float] = None
    #: URL of image showing the map of the location
    image_url: Optional[str] = field(default=None, init=False)
    #: Width of the image
    image_width: Optional[int] = field(default=None, init=False)
    #: Height of the image
    image_height: Optional[int] = field(default=None, init=False)
    #: URL to Bing maps with the location
    url: Optional[str] = field(default=None, init=False)
    # Address of the location
    address: Optional[str] = field(default=None)

    # Put here for backwards compatibility, so that the init argument order is preserved
    uid: Optional[int] = None

    @classmethod
    def _from_graphql(cls, data)-> LocationAttachment:
        url = data.get("url")
        address = _util.get_url_parameter(_util.get_url_parameter(url, "u"), "where1")
        try:
            latitude, longitude = [float(x) for x in address.split(", ")]
            address = None
        except ValueError:
            latitude, longitude = None, None
        rtn = cls(
            uid=int(data["deduplication_key"]),
            latitude=latitude,
            longitude=longitude,
            address=address,
        )
        media = data.get("media")
        if media and media.get("image"):
            image = media["image"]
            rtn.image_url = image.get("uri")
            rtn.image_width = image.get("width")
            rtn.image_height = image.get("height")
        rtn.url = url
        return rtn


@dataclass(eq=False, init=False)
class LiveLocationAttachment(LocationAttachment):
    """Represents a live user location."""

    #: Name of the location
    name: Optional[str] = None
    #: Timestamp when live location expires
    expiration_time: Optional[int] = None
    #: True if live location is expired
    is_expired: Optional[bool] = None

    def __init__(self, name=None, expiration_time=None, is_expired=None, **kwargs):
        super(LiveLocationAttachment, self).__init__(**kwargs)
        self.expiration_time = expiration_time
        self.is_expired = is_expired

    @classmethod
    def _from_pull(cls, data)-> LiveLocationAttachment:
        return cls(
            uid=data["id"],
            latitude=data["coordinate"]["latitude"] / (10 ** 8)
            if not data.get("stopReason")
            else None,
            longitude=data["coordinate"]["longitude"] / (10 ** 8)
            if not data.get("stopReason")
            else None,
            name=data.get("locationTitle"),
            expiration_time=data["expirationTime"],
            is_expired=bool(data.get("stopReason")),
        )

    @classmethod
    def _from_graphql(cls, data)-> LiveLocationAttachment:
        target = data["target"]
        rtn = cls(
            uid=int(target["live_location_id"]),
            latitude=target["coordinate"]["latitude"]
            if target.get("coordinate")
            else None,
            longitude=target["coordinate"]["longitude"]
            if target.get("coordinate")
            else None,
            name=data["title_with_entities"]["text"],
            expiration_time=target.get("expiration_time"),
            is_expired=target.get("is_expired"),
        )
        media = data.get("media")
        if media and media.get("image"):
            image = media["image"]
            rtn.image_url = image.get("uri")
            rtn.image_width = image.get("width")
            rtn.image_height = image.get("height")
        rtn.url = data.get("url")
        return rtn



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Optional, List
from .. import _util


@dataclass(eq=False)
class Attachment:
    """Represents a Facebook attachment."""
    #: The attachment ID
    uid: Optional[int] = field(default=None)

@dataclass(eq=False)
class UnsentMessage(Attachment):
    """Represents an unsent message attachment."""


@dataclass(eq=False)
class ShareAttachment(Attachment):
    """Represents a shared item (e.g. URL) attachment."""

    #: ID of the author of the shared post
    author: Optional[str] = field(default=None)
    #: Target URL
    url: Optional[str] = field(default=None)
    #: Original URL if Facebook redirects the URL
    original_url: Optional[str] = field(default=None)
    #: Title of the attachment
    title: Optional[str] = field(default=None)
    #: Description of the attachment
    description: Optional[str] = field(default=None)
    #: Name of the source
    source: Optional[str] = field(default=None)
    #: URL of the attachment image
    image_url: Optional[str] = field(default=None)
    #: URL of the original image if Facebook uses ``safe_image``
    original_image_url: Optional[str] = field(default=None)
    #: Width of the image
    image_width: Optional[int] = field(default=None)
    #: Height of the image
    image_height: Optional[int] = field(default=None)
    #: List of additional attachments
    attachments: List = field(default_factory=list)

    # Put here for backwards compatibility, so that the init argument order is preservd
    uid: Optional[int] = field(default=None)

    def __post_init__(self):
        if self.attachments is None:
            self.attachments = []



    @classmethod
    def _from_graphql(cls, data)-> ShareAttachment:
        from . import _file

        url = data.get("url")
        rtn = cls(
            uid=data.get("deduplication_key"),
            author=data["target"]["actors"][0]["id"]
            if data["target"].get("actors")
            else None,
            url=url,
            original_url=_util.get_url_parameter(url, "u")
            if "/l.php?u=" in url
            else url,
            title=data["title_with_entities"].get("text"),
            description=data["description"].get("text")
            if data.get("description")
            else None,
            source=data["source"].get("text") if data.get("source") else None,
            attachments=[
                _file.graphql_to_subattachment(attachment)
                for attachment in data.get("subattachments")
            ],
        )
        media = data.get("media")
        if media and media.get("image"):
            image = media["image"]
            rtn.image_url = image.get("uri")
            rtn.original_image_url = (
                _util.get_url_parameter(rtn.image_url, "url")
                if "/safe_image.php" in rtn.image_url #type: ignore
                else rtn.image_url
            )
            rtn.image_width = image.get("width")
            rtn.image_height = image.get("height")
        return rtn



from __future__ import annotations

from dataclasses import field, dataclass
import json
from string import Formatter
from typing import Any, List, Dict, Optional, Union
from ._thread import ThreadType
from . import _attachment, _location, _file, _sticker
from .. import _util
from ._quick_reply import QuickReplyText, QuickReplyEmail, QuickReplyLocation, QuickReplyPhoneNumber, graphql_to_quick_reply
from ._core import CustomEnum

QuickReplies = Union[
    QuickReplyPhoneNumber,
    QuickReplyLocation,
    QuickReplyEmail,
    QuickReplyText
]


class EmojiSize(CustomEnum):
    """
    Specifies the size of an emoji sent in a message.

    Attributes:
        LARGE: Represents a large emoji.
        MEDIUM: Represents a medium emoji.
        SMALL: Represents a small emoji.
    """

    LARGE = "369239383222810"
    MEDIUM = "369239343222814"
    SMALL = "369239263222822"

    @classmethod
    def _from_tags(cls, tags: List[str])-> Optional[EmojiSize]:

        string_to_emojisize: Dict = {
            "large": cls.LARGE,
            "medium": cls.MEDIUM,
            "small": cls.SMALL,
            "l": cls.LARGE,
            "m": cls.MEDIUM,
            "s": cls.SMALL,
        }
        for tag in tags or ():
            data = tag.split(":", 1)
            if len(data) > 1 and data[0] == "hot_emoji_size": 
                return string_to_emojisize[data[1]]
                
        return None


class MessageReaction(CustomEnum):
    """
    Specifies the type of reaction to a message.

    Attributes:
        HEART: Represents a heart reaction.
        LOVE: Represents a love reaction.
        SMILE: Represents a smile reaction.
        WOW: Represents a wow reaction.
        SAD: Represents a sad reaction.
        ANGRY: Represents an angry reaction.
        YES: Represents a thumbs-up reaction.
        NO: Represents a thumbs-down reaction.
    """

    HEART = "❤"
    LOVE = "😍"
    SMILE = "😆"
    WOW = "😮"
    SAD = "😢"
    ANGRY = "😠"
    YES = "👍"
    NO = "👎"


@dataclass(eq=False)
class Mention:
    """
    Represents a mention in a message.

    Attributes:
        thread_id: The ID of the thread being mentioned.
        offset: The character offset where the mention starts.
        length: The length of the mention.
    """
    thread_id: str = field()
    offset: int = field(default=0)
    length: int = field(default=10)


@dataclass(eq=False)
class MessageFunc:
    """
    Base class for message-related actions such as reacting, replying, and unsending.

    Attributes:
        uid: The unique identifier for the message.
        client: Reference to the Client instance.
        author: ID of the message sender.
        thread_id: ID of the thread where the message belongs.
        thread_type: Type of the thread (e.g., USER, GROUP).
    """

    uid: str = field(init=False)

    client: Any = field(init=False)

    author: str = field(init=False)

    thread_id: str = field(init=False)

    thread_type: Any = field(init=False)

    @property
    def state(self):
        return self.client._state



    async def react(self, reaction: str):
        """
        React to a message.

        Args:
            reaction: The reaction to apply, or None to remove the reaction.

        Raises:
            RuntimeError: If the reaction action fails.
        """
        data = {
        "action": "ADD_REACTION" if reaction else "REMOVE_REACTION",
        "client_mutation_id": "1",
        "actor_id": self.client.uid,
        "message_id": self.uid,
        "reaction": reaction,
        }
        data = {
            "doc_id": 1491398900900362,
            "variables": _util.json_minimal({"data": data}),
        }
        try:
            await self.state._payload_post("/webgraphql/mutation", data)
        except Exception as e:
            raise RuntimeError("failed reacting: ", e)


    async def reply(self, message):
        """
        Reply to a message.

        Args:
            message: The text of the reply.
        """
        data = {}
        if self.thread_type == ThreadType.USER:
            data = {
                "other_user_fbid": self.thread_id,
                "action_type": "ma-type:user-generated-message",
                "body": message,
                "replied_to_message_id": self.uid
            }

            await self.state.do_send_request(data)

        elif self.thread_type == ThreadType.GROUP:
            data = {
                "thread_fbid": self.thread_id,
                "action_type": "ma-type:user-generated-message",
                "body": message,
                "replied_to_message_id": self.uid
                }
            await self.state.do_send_request(data)


    async def unsend(self):
        """
        Unsend (delete) the message if the author is the current user.

        Prints:
            Success message upon completion.
        """
        if self.author == self.state.user_id:
            data = {"message_id": self.uid}
            await self.state._payload_post("/messaging/unsend_message/?dpr=1", data)
            print("message deleted succesfully")


@dataclass(eq=False)
class Message(MessageFunc):
    """
    Represents a Facebook message with its details.
    """
    #: The text of the message
    text: Optional[str] = field(default=None)
    #: A list of Mention object
    mentions: List[Mention] = field(default_factory=list)
    #: The size of a sent emoji
    emoji_size: Optional[EmojiSize] = field(default=None)
    #: The ID of the message
    uid: str = field(init=False)
    #: The ID of the message sender
    author: str = field(init=False)
    #: The timestamp when the message was sent
    timestamp: Optional[int] = field(default=None, init=False)
    #: 
    is_read: Optional[bool] = field(default=None, init=False)
    #: A list of people IDs who read the message, works only with :func:`fbchat.Client.fetchThreadMessages`
    read_by: List[str] = field(default_factory=list, init=False)
    #: A dictionary with user's IDs as keys, and their :class:`MessageReaction` as values
    reactions: Dict[str, MessageReaction] = field(default_factory=dict, init=False)
    #: A :class:`Sticker`
    sticker: Optional[_sticker.Sticker] = field(default=None)
    #: A list of attachments
    attachments: List[_attachment.Attachment] = field(default_factory=list)
    #: A list of :class:`QuickReply`
    quick_replies: Optional[List[QuickReplies]] = field(default_factory=list)
    #: Whether the message is unsent (deleted for everyone)
    unsent: bool = field(default=False, init=False)
    #: Message ID you want to reply to
    reply_to_id: Optional[str] = field(default=None)
    #: Replied message
    replied_to: Optional["Message"] = field(default=None, init=False)
    #: Whether the message was forwarded
    forwarded: bool = field(default=False, init=False)



    def __post_init__(self):
        # Converter logic for mentions
        if self.mentions is None:
            self.mentions = []
        if self.attachments is None:
            self.attachments = []
        if self.quick_replies is None:
            self.quick_replies = []
        if isinstance(self.emoji_size, str):
            self.emoji_size = EmojiSize(self.emoji_size)
    


    @classmethod
    def formatMentions(cls, text, *args, **kwargs):
        # """Like `str.format`, but takes tuples with a thread id and text instead.
        #
        # Return a `Message` object, with the formatted string and relevant mentions.
        #
        # >>> Message.formatMentions("Hey {!r}! My name is {}", ("1234", "Peter"), ("4321", "Michael"))
        # <Message (None): "Hey 'Peter'! My name is Michael", mentions=[<Mention 1234: offset=4 length=7>, <Mention 4321: offset=24 length=7>] emoji_size=None attachments=[]>
        #
        # >>> Message.formatMentions("Hey {p}! My name is {}", ("1234", "Michael"), p=("4321", "Peter"))
        # <Message (None): 'Hey Peter! My name is Michael', mentions=[<Mention 4321: offset=4 length=5>, <Mention 1234: offset=22 length=7>] emoji_size=None attachments=[]>
        # """
        result = ""
        mentions = list()
        offset = 0
        f = Formatter()
        field_names = [field_name[1] for field_name in f.parse(text)]
        automatic = "" in field_names
        i = 0

        for (literal_text, field_name, format_spec, conversion) in f.parse(text):
            offset += len(literal_text)
            result += literal_text

            if field_name is None:
                continue

            if field_name == "":
                field_name = str(i)
                i += 1
            elif automatic and field_name.isdigit():
                raise ValueError(
                    "cannot switch from automatic field numbering to manual field specification"
                )

            thread_id, name = f.get_field(field_name, args, kwargs)[0]

            if format_spec:
                name = f.format_field(name, format_spec)
            if conversion:
                name = f.convert_field(name, conversion)

            result += name
            mentions.append(
                Mention(thread_id=thread_id, offset=offset, length=len(name))
            )
            offset += len(name)

        message = cls(text=result, mentions=mentions)
        return message

    @staticmethod
    def _get_forwarded_from_tags(tags):
        if tags is None:
            return False
        return any(map(lambda tag: "forward" in tag or "copy" in tag, tags))

    def _to_send_data(self)-> Dict:
        data = {}

        if self.text or self.sticker or self.emoji_size:
            data["action_type"] = "ma-type:user-generated-message"

        if self.text:
            data["body"] = self.text

        for i, mention in enumerate(self.mentions):
            data["profile_xmd[{}][id]".format(i)] = mention.thread_id
            data["profile_xmd[{}][offset]".format(i)] = mention.offset
            data["profile_xmd[{}][length]".format(i)] = mention.length
            data["profile_xmd[{}][type]".format(i)] = "p"

        if self.emoji_size:
            if self.text:
                data["tags[0]"] = "hot_emoji_size:" + self.emoji_size.name.lower()
            else:
                data["sticker_id"] = self.emoji_size.value

        if self.sticker:
            data["sticker_id"] = self.sticker.uid

        if self.quick_replies:
            xmd = {"quick_replies": []}
            for quick_reply in self.quick_replies:
                # TODO: Move this to `_quick_reply.py`
                q = dict()
                q["content_type"] = quick_reply._type
                q["payload"] = quick_reply.payload
                q["external_payload"] = quick_reply.external_payload
                q["data"] = quick_reply.data
                if quick_reply.is_response:
                    q["ignore_for_webhook"] = False
                if isinstance(quick_reply, QuickReplyText):
                    q["title"] = quick_reply.title
                if not isinstance(quick_reply, QuickReplyLocation):
                    q["image_url"] = quick_reply.image_url
                xmd["quick_replies"].append(q)
            if len(self.quick_replies) == 1 and self.quick_replies[0].is_response:
                xmd["quick_replies"] = xmd["quick_replies"][0]
            data["platform_xmd"] = json.dumps(xmd)

        if self.reply_to_id:
            data["replied_to_message_id"] = self.reply_to_id

        return data

    @classmethod
    def _from_graphql(cls, data)-> Message:
        if data is None:
            return
        if data.get("message_sender") is None:
            data["message_sender"] = {}
        if data.get("message") is None:
            data["message"] = {}
        tags = data.get("tags_list")
        rtn = cls(
            text=data["message"].get("text"),
            mentions=[
                Mention(
                    m.get("entity", {}).get("id"),
                    offset=m.get("offset"),
                    length=m.get("length"),
                )
                for m in data["message"].get("ranges") or ()
            ],
            emoji_size=EmojiSize._from_tags(tags),
            sticker=_sticker.Sticker._from_graphql(data.get("sticker")),
        )
        rtn.forwarded = cls._get_forwarded_from_tags(tags)
        rtn.uid = str(data["message_id"])
        rtn.author = str(data["message_sender"]["id"])
        rtn.timestamp = data.get("timestamp_precise")
        rtn.unsent = False
        if data.get("unread") is not None:
            rtn.is_read = not data["unread"]
        rtn.reactions = {
            str(r["user"]["id"]): MessageReaction._extend_if_invalid(r["reaction"])
            for r in data["message_reactions"]
        }
        if data.get("blob_attachments") is not None:
            rtn.attachments = [
                _file.graphql_to_attachment(attachment)
                for attachment in data["blob_attachments"]
            ]
        if data.get("platform_xmd_encoded"):
            quick_replies = json.loads(data["platform_xmd_encoded"]).get(
                "quick_replies"
            )
            if isinstance(quick_replies, list):
                rtn.quick_replies = [
                    graphql_to_quick_reply(q) for q in quick_replies
                ]
            elif isinstance(quick_replies, dict):
                rtn.quick_replies = [
                    graphql_to_quick_reply(quick_replies, is_response=True)
                ]
        if data.get("extensible_attachment") is not None:
            attachment = graphql_to_extensible_attachment(data["extensible_attachment"])
            if isinstance(attachment, _attachment.UnsentMessage):
                rtn.unsent = True
            elif attachment:
                rtn.attachments.append(attachment)
        if data.get("replied_to_message") is not None:
            rtn.replied_to = cls._from_graphql(data["replied_to_message"]["message"])
            print("this is replied_to: ", rtn.replied_to)
            if rtn.replied_to is not None:
                rtn.reply_to_id = rtn.replied_to.uid
        return rtn

    @classmethod
    def _from_reply(cls, data, client, thread_id, thread_type)-> Message:
        tags = data["messageMetadata"].get("tags")
        rtn = cls(
            text=data.get("body"),
            mentions=[
                Mention(m.get("i"), offset=m.get("o"), length=m.get("l"))
                for m in json.loads(data.get("data", {}).get("prng", "[]"))
            ],
            emoji_size=EmojiSize._from_tags(tags),
        )
        metadata = data.get("messageMetadata", {})
        rtn.forwarded = cls._get_forwarded_from_tags(tags)
        rtn.uid = metadata.get("messageId")
        rtn.author = str(metadata.get("actorFbId"))
        rtn.timestamp = metadata.get("timestamp")
        rtn.client = client
        rtn.thread_id = thread_id
        rtn.thread_type = thread_type
        rtn.unsent = False
        if data.get("data", {}).get("platform_xmd"):
            quick_replies = json.loads(data["data"]["platform_xmd"]).get(
                "quick_replies"
            )
            if isinstance(quick_replies, list):
                if quick_replies is not None:
                    rtn.quick_replies = [
                    graphql_to_quick_reply(q) for q in quick_replies
                ]
            elif isinstance(quick_replies, dict):
                rtn.quick_replies = [
                    graphql_to_quick_reply(quick_replies, is_response=True)
                ]
        if data.get("attachments") is not None:
            for attachment in data["attachments"]:
                attachment = json.loads(attachment["mercuryJSON"])
                if attachment.get("blob_attachment"):
                    rtn.attachments.append(
                        _file.graphql_to_attachment(attachment["blob_attachment"])
                    )
                if attachment.get("extensible_attachment"):
                    extensible_attachment = graphql_to_extensible_attachment(
                        attachment["extensible_attachment"]
                    )
                    if isinstance(extensible_attachment, _attachment.UnsentMessage):
                        rtn.unsent = True
                    else:
                        rtn.attachments.append(extensible_attachment)
                if attachment.get("sticker_attachment"):
                    rtn.sticker = _sticker.Sticker._from_graphql(
                        attachment["sticker_attachment"]
                    )
        return rtn

    @classmethod
    def _from_pull(cls, data, mid: str , tags: List[str], author: str, timestamp: int , client, thread_id ,thread_type)-> Message:
        rtn = cls(text=data.get("body"))
        rtn.uid = mid
        rtn.author = author
        rtn.timestamp = timestamp
        rtn.client = client
        rtn.thread_id = thread_id
        rtn.thread_type = thread_type

        if data.get("data") and data["data"].get("prng"):
            try:
                rtn.mentions = [
                    Mention(
                        str(mention.get("i")),
                        offset=mention.get("o"),
                        length=mention.get("l"),
                    )
                    for mention in _util.parse_json(data["data"]["prng"])
                ]
            except Exception:
                print("An exception occured while reading attachments")

        if data.get("attachments"):
            try:
                for a in data["attachments"]:
                    mercury = a["mercury"]
                    if mercury.get("blob_attachment"):
                        image_metadata = a.get("imageMetadata", {})
                        attach_type = mercury["blob_attachment"]["__typename"]
                        attachment = _file.graphql_to_attachment(
                            mercury["blob_attachment"]
                        )

                        if attach_type in [
                            "MessageFile",
                            "MessageVideo",
                            "MessageAudio",
                        ]:
                            # TODO: Add more data here for audio files
                            print("this is attachment.size: ", int(a["fileSize"]))
                            attachment.size = int(a["fileSize"])
                        rtn.attachments.append(attachment)

                    elif mercury.get("sticker_attachment"):
                        rtn.sticker = _sticker.Sticker._from_graphql(
                            mercury["sticker_attachment"]
                        )

                    elif mercury.get("extensible_attachment"):
                        attachment = graphql_to_extensible_attachment(
                            mercury["extensible_attachment"]
                        )
                        if isinstance(attachment, _attachment.UnsentMessage):
                            rtn.unsent = True
                        elif attachment:
                            rtn.attachments.append(attachment)

            except Exception:
                print(
                    "An exception occured while reading attachments: {}".format(
                        data["attachments"]
                    )
                )

        rtn.emoji_size = EmojiSize._from_tags(tags)
        rtn.forwarded = cls._get_forwarded_from_tags(tags)
        return rtn


def graphql_to_extensible_attachment(data):
    story = data.get("story_attachment")
    if not story:
        return None

    target = story.get("target")
    if not target:
        return _attachment.UnsentMessage(uid=data.get("legacy_attachment_id"))

    _type = target["__typename"]
    if _type == "MessageLocation":
        return _location.LocationAttachment._from_graphql(story)
    elif _type == "MessageLiveLocation":
        return _location.LiveLocationAttachment._from_graphql(story)
    elif _type in ["ExternalUrl", "Story"]:
        return _attachment.ShareAttachment._from_graphql(story)

    return None



from __future__ import annotations
from dataclasses import dataclass
from typing import List, Optional
from . import _plan
from ._thread import ThreadType, Thread


@dataclass(eq=False)
class Page(Thread):
    """Represents a Facebook page. Inherits `Thread`."""

    #: The page's custom URL
    url: Optional[str] = None
    #: The name of the page's location city
    city: Optional[str] = None
    #: Amount of likes the page has
    likes: Optional[int] = None
    #: Some extra information about the page
    sub_title: Optional[str] = None
    #: The page's category
    category: Optional[List] = None

    def __post_init__(self):
        self.type = ThreadType.PAGE

    @classmethod
    def _from_graphql(cls, data)-> Page:
        if data.get("profile_picture") is None:
            data["profile_picture"] = {}
        if data.get("city") is None:
            data["city"] = {}
        plan = None
        if data.get("event_reminders") and data["event_reminders"].get("nodes"):
            plan = _plan.Plan._from_graphql(data["event_reminders"]["nodes"][0])

        return cls(
            data["id"],
            url=data.get("url"),
            city=data.get("city").get("name"),
            category=data.get("category_type"),
            photo=data["profile_picture"].get("uri"),
            name=data.get("name"),
            message_count=data.get("messages_count"),
            plan=plan,
        )



import random
import time
import json
from typing import List, Set, Tuple
import aiohttp
from os.path import basename 
from mimetypes import guess_type
from contextlib import contextmanager
from urllib.parse import urlparse, parse_qs
from yarl import URL


def save_cookies_to_session(session: aiohttp.ClientSession, cookies):
    """
    Save cookies from a list to an aiohttp.ClientSession.
    
    Args:
        session (aiohttp.ClientSession): The client session to update.
        cookies (list): A list of cookies in the format:
            [{'key': 'name', 'value': 'value'}, ...]
    """
        # Add modified cookie for Messenger domain
    messenger_domain = 'www.messenger.com'
    session.cookie_jar.update_cookies(
        cookies,
        response_url=URL(f"https://{messenger_domain}")
            )


def now_time()-> int:
    "returns current time"
    return int(time.time() * 1000)



def to_json(content)-> dict:
    """Removes `for(;;);` (and other cruft) that preceeds JSON responses."""
    try:
        content = content[content.index("{"):]
        return json.loads(content)
    except Exception as e:
        raise ValueError("Failed parsing Json", e)


def digitToChar(digit):
    if digit < 10:
        return str(digit)
    return chr(ord("a") + digit - 10)


def str_base(number, base):
    if number < 0:
        return "-" + str_base(-number, base)
    (d, m) = divmod(number, base)
    if d > 0:
        return str_base(d, base) + digitToChar(m)
    return digitToChar(m)



def decimal_to_base36(decimal_num)->str:
    if not isinstance(decimal_num, int):
        raise ValueError("Input must be an integer.")
    if decimal_num == 0:
        return "0"
    base36_chars = "0123456789abcdefghijklmnopqrstuvwxyz"
    result = []
    while decimal_num:
        decimal_num, remainder = divmod(decimal_num, 36)
        result.append(base36_chars[remainder])

    return ''.join(result[::-1])



def check_http_code(status_code: int):
    msg = f"Error when sending request: Got {status_code} response"
    if status_code == 404:
        print(f"{msg}\nThis is either because you specified an invalid URL, or because you provided an invalid id (Facebook usually requires integer ids")
    elif 400 <=status_code < 600:
        print(msg)


async def check_request(response)-> str:
    check_http_code(response.status)
    content = await response.text()
    if content is None or not len(content):
        raise ValueError("Error When sending request: Got empty response")
    return content

def prefix_url(url: str)-> URL:
    """Adds https://www.facebook.com"""
    if str(url).startswith("/"):
        return URL("https://www.facebook.com" + url)
    return URL(url)


def generateMessageID(client_id=None)-> str:
      k = now_time()
      lame = int(random.random() * 4294967295)
      return f"<{k}:{lame}-{client_id}@mail.projektitan.com>"


def getSignatureID()->str:
    return hex(int(random.random() * 2147483648))


def generateOfflineThreadingID()-> str:
    ret = now_time()
    value = int(random.random() * 4294967295)
    string = ("0000000000000000000000" + format(value, "b"))[-22:]
    msgs = format(ret, "b") + string
    return str(int(msgs, 2))

def require_list(list_)-> Set[str]:
      if isinstance(list_, list):
         return set(list_)
      else:
         return set([list_])

def handle_payload_error(j):
    if "error" not in j:
        return
    error_code = j["error"]
    if error_code == 1357001:
        print("FBChat not logged in")
    elif error_code == 1357004:
        print("Error: please refresh your cookies!")
    elif error_code in (1357031, 1545010, 1545003):
        print("Error: Invalid Parameters")
    else:
        print("FBchatError")

def handle_graphql_error(response):
    """Handles Erros"""
    errors = []
    if response.get("error"):
        errors = [response["error"]]
    elif "errors" in response:
        errors = response["error"]
    if errors:
        error = errors[0]
        print(f"Error:  error-code: {error.get('code')} error-message: {error.get('message')}")
 

def mimetype_to_key(mimetype) -> str:
    """This function is used when sending files, images"""
    if not mimetype:
        return "file_id"  # Default for unknown mimetypes
    
    mimetype = mimetype.strip()  # Remove leading/trailing whitespace
    if mimetype == "image/gif":
        return "gif_id"
    
    parts = mimetype.split("/")
    if len(parts) == 2 and parts[0] in ["video", "image", "audio"]:
        return f"{parts[0]}_id"
    
    return "file_id"  # Default for unsupported types


def get_jsmods_require(j, index):
    if j.get("jsmods") and j["jsmods"].get("require"):
        try:
            return j["jsmods"]["require"][0][index][0]
        except (KeyError, IndexError) as e:
            print(
                "Error when getting jsmods_require: {}. Facebook might have changed protocol".format(j)
            )
    return None


async def get_files_from_urls(file_urls)-> List[Tuple[str, bytes, str]]:
    files = []
    async with aiohttp.ClientSession() as session:
        for file_url in file_urls:
            async with session.get(file_url) as response:
                if response.status != 200:
                    raise aiohttp.ClientResponseError(
                        response.request_info,
                        response.history,
                        status=response.status,
                        message=f"Failed to fetch {file_url}",
                    )
                file_name = basename(file_url).split("?")[0].split("#")[0]
                content_type = response.headers.get("Content-Type") or guess_type(file_name)[0]
                content = await response.read()  # Read the content as bytes
                files.append(
                    (
                        file_name,
                        content,  # Use bytes, not StreamReader
                        content_type,
                    )
                )
    return files


@contextmanager
def get_files_from_paths(filenames):
    files = []
    for filename in filenames:
        files.append(
            (basename(filename), open(filename, "rb"), guess_type(filename)[0])
        )
    yield files
    for fn, fp, ft in files:
        fp.close()


def get_url_parameters(url, *args):
    params = parse_qs(urlparse(url).query)
    return [params[arg][0] for arg in args if params.get(arg)]


def get_url_parameter(url, param):
    return get_url_parameters(url, param)[0]


def get_cookie(json_path: str) -> dict:
    """Get ans Format cookies before passing to session"""
    with open(json_path, "r") as f:
        data = json.load(f)
    cookies = {}
    for cookie in data:
        cookies[cookie["key"]] = cookie["value"]
    return cookies



from __future__ import annotations

import aiomqtt
import paho.mqtt.client as mqtt
import random
import aiohttp

from dataclasses import dataclass, field
from typing import Any
from yarl import URL

from . import _graphql, _util
from aiomqtt.exceptions import MqttConnectError
from .n_state import State


@dataclass(slots=True)
class Mqtt:
    _state: State = field()
    _mqttClient: aiomqtt.Client = field()
    _chat_on: bool = field()
    _foreground: bool = field()
    _sequence_id: int = field()
    _sync_token: Any = field(default=None)
    _HOST = "edge-chat.facebook.com"


    @classmethod
    async def connect(cls, state: State, chat_on: bool, foreground: bool)-> Mqtt:

        mqttClint = aiomqtt.Client(
            hostname=cls._HOST,
            identifier="mqttwsclient",
            clean_session=True,
            protocol=aiomqtt.ProtocolVersion.V31,
            transport="websockets",
            port=443,
            keepalive=60
        )
        sequence_id = await cls._fetch_sequence_id(state)
        # needed for websockets
        mqttClint._client.tls_set()
        # creating class instance 
        self = cls(
            _state=state,
            _mqttClient=mqttClint,
            _chat_on=chat_on,
            _foreground=foreground,
            _sequence_id=sequence_id
        )
    
        
        self._configure_connect_options()
        await mqttClint.__aenter__()
        if self._mqttClient._client.is_connected():
            await self._messenger_queue_publish()
        return self

    @staticmethod
    async def _fetch_sequence_id(state)-> int:
        """Fetch sequence ID."""
        params = {
            "limit": 1,
            "tags": ["INBOX"],
            "before": None,
            "includeDeliveryReceipts": False,
            "includeSeqID": True,
            }
        (j,) = await state._graphql_requests(_graphql.from_doc_id("1349387578499440", params))
        sequence_id = j["viewer"]["message_threads"]["sync_sequence_id"] #type: ignore
        return int(sequence_id)



    def _configure_connect_options(self)-> None:
        session_id = generate_session_id()
        topics = [
            "/t_ms",
            "/thread_typing",
            "/orca_typing_notifications",
            "/orca_presence",
            "/legacy_web", "/br_sr", "/sr_res",
            "ls_resp",
            "/webrtc",
            "/onevc",
            "/notify_disconnect",
            "/inbox",
            "/mercury",
            "/messaging_events",
            "/orca_message_notifications",
            "/pp",
            "/webrtc_response",
        ]
        username = {
            "u": self._state.user_id,
            "s": session_id,
            "chat_on": self._chat_on,
            "fg": self._foreground,
            "d": self._state._client_id,
            "aid": 219994525426954,
            "st": topics,
            "pm": [],
            "cp": 3,
            "ecp": 10,
            "ct": "websocket",
            "mqtt_sid": "",
            "dc": "",
            "no_auto_fg": True,
            "gas": None,
            "pack": [],
        }
        self._mqttClient._client.username_pw_set(_util.json_minimal(username))

        headers = {
            "Cookie": get_cookie_header(
                self._state._session, "https://edge-chat.facebook.com/chat"
            ),
            "User-Agent": self._state._session.headers["User-Agent"],
            "Origin": "https://www.facebook.com",
            "Host": self._HOST
        }
        self._mqttClient._client.ws_set_options(
            path="/chat?sid={}".format(session_id), headers=headers
        )
        
    
    async def _messenger_queue_publish(self)-> None:
        # configure receiving messages.
        payload = {
            "sync_api_version": 10,
            "max_deltas_able_to_process": 1000,
            "delta_batch_size": 500,
            "encoding": "JSON",
            "entity_fbid": self._state.user_id,
        }

        if self._sync_token is None:
            topic = "/messenger_sync_create_queue"
            payload["initial_titan_sequence_id"] = str(self._sequence_id)
            payload["device_params"] = None
        else:
            topic = "/messenger_sync_get_diffs"
            payload["last_seq_id"] = str(self._sequence_id)
            payload["sync_token"] = self._sync_token

        await self._mqttClient.publish(topic, _util.json_minimal(payload), qos=1)

    def _on_connect_client(self, client, userdata, flags, reason_code) -> None:
        """Called when we receive a CONNACK message from the broker."""
        if self._mqttClient._connected.done():
            pass
        if reason_code == mqtt.CONNACK_ACCEPTED:
            self._mqttClient._connected.set_result(None)
        else:
            # We received a negative CONNACK response
            self._mqttClient._connected.set_exception(MqttConnectError(reason_code))


    async def disconnect(self):
        await self._mqttClient.__aexit__(None, None, None)

    async def set_foreground(self, value):
        payload = _util.json_minimal({"foreground": self._chat_on})
        await self._mqttClient.publish("/foreground_state", payload=payload, qos=1)
        self._foreground = value

    async def set_chat_on(self, value):
        data = {"make_user_available_when_in_foreground": True}
        payload = _util.json_minimal(data)
        await self._mqttClient.publish("/set_client_settings", payload=payload, qos=1)
        self._chat_on = value

def generate_session_id()-> int:
    return random.randint(1, 2 ** 53)


def get_cookie_header(session: aiohttp.ClientSession, url: str) -> str:
    return session.cookie_jar.filter_cookies(URL(url)).output(header="", sep=";").lstrip()




from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional
from ._core import CustomEnum
from . import _plan
from ._thread import Thread, ThreadType


GENDERS = {
    # For standard requests
    0: "unknown",
    1: "female_singular",
    2: "male_singular",
    3: "female_singular_guess",
    4: "male_singular_guess",
    5: "mixed",
    6: "neuter_singular",
    7: "unknown_singular",
    8: "female_plural",
    9: "male_plural",
    10: "neuter_plural",
    11: "unknown_plural",
    # For graphql requests
    "UNKNOWN": "unknown",
    "FEMALE": "female_singular",
    "MALE": "male_singular",
    # '': 'female_singular_guess',
    # '': 'male_singular_guess',
    # '': 'mixed',
    "NEUTER": "neuter_singular",
    # '': 'unknown_singular',
    # '': 'female_plural',
    # '': 'male_plural',
    # '': 'neuter_plural',
    # '': 'unknown_plural',
}


class TypingStatus(CustomEnum):
    """Used to specify whether the user is typing or has stopped typing."""

    STOPPED = 0
    TYPING = 1


@dataclass(eq=False)
class User(Thread):
    """Represents a Facebook user. Inherits `Thread`."""

    #: The profile URL
    url: Optional[str] = None
    #: The users first name
    first_name: Optional[str] = None
    #: The users last name
    last_name: Optional[str] = None
    #: Whether the user and the client are friends
    is_friend: Optional[bool] = None
    #: The user's gender
    gender: Optional[str] = None
    #: From 0 to 1. How close the client is to the user
    affinity: Optional[int] = None
    #: The user's nickname
    nickname: Optional[str] = None
    #: The clients nickname, as seen by the user
    own_nickname: Optional[str] = None
    #: A :class:`ThreadColor`. The message color
    color: Any = None
    #: The default emoji
    emoji: Any = None


    def __post_init__(self):
        self.type = ThreadType.USER
    @classmethod
    def _from_graphql(cls, data)-> User:
        if data.get("profile_picture") is None:
            data["profile_picture"] = {}
        c_info = cls._parse_customization_info(data)
        plan = None
        if data.get("event_reminders") and data["event_reminders"].get("nodes"):
            plan = _plan.Plan._from_graphql(data["event_reminders"]["nodes"][0])

        return cls(
            data["id"],
            url=data.get("url"),
            first_name=data.get("first_name"),
            last_name=data.get("last_name"),
            is_friend=data.get("is_viewer_friend"),
            gender=GENDERS.get(data.get("gender")),
            affinity=data.get("affinity"),
            nickname=c_info.get("nickname"),
            color=c_info.get("color"),
            emoji=c_info.get("emoji"),
            own_nickname=c_info.get("own_nickname"),
            photo=data["profile_picture"].get("uri"),
            name=data.get("name"),
            message_count=data.get("messages_count"),
            plan=plan,
        )

    @classmethod
    def _from_thread_fetch(cls, data)-> User:
        if data.get("big_image_src") is None:
            data["big_image_src"] = {}
        c_info = cls._parse_customization_info(data)
        participants = [
            node["messaging_actor"] for node in data["all_participants"]["nodes"]
        ]
        user = next(
            p for p in participants if p["id"] == data["thread_key"]["other_user_id"]
        )
        last_message_timestamp = None
        if "last_message" in data:
            last_message_timestamp = data["last_message"]["nodes"][0][
                "timestamp_precise"
            ]

        first_name = user.get("short_name")
        if first_name is None:
            last_name = None
        else:
            last_name = user.get("name").split(first_name, 1).pop().strip()

        plan = None
        if data.get("event_reminders") and data["event_reminders"].get("nodes"):
            plan = _plan.Plan._from_graphql(data["event_reminders"]["nodes"][0])

        return cls(
            user["id"],
            url=user.get("url"),
            name=user.get("name"),
            first_name=first_name,
            last_name=last_name,
            is_friend=user.get("is_viewer_friend"),
            gender=GENDERS.get(user.get("gender")),
            affinity=user.get("affinity"),
            nickname=c_info.get("nickname"),
            color=c_info.get("color"),
            emoji=c_info.get("emoji"),
            own_nickname=c_info.get("own_nickname"),
            photo=user["big_image_src"].get("uri"),
            message_count=data.get("messages_count"),
            last_message_timestamp=last_message_timestamp,
            plan=plan,
        )

    @classmethod
    def _from_all_fetch(cls, data)-> User:
        return cls(
            data["id"],
            first_name=data.get("firstName"),
            url=data.get("uri"),
            photo=data.get("thumbSrc"),
            name=data.get("name"),
            is_friend=data.get("is_friend"),
            gender=GENDERS.get(data.get("gender")),
        )


@dataclass(eq=False)
class ActiveStatus:
    """Represents Facebook Active Status"""
    #: Whether the user is active now
    active: bool
    #: Timestamp when the user was last active
    last_active: Optional[int] = None
    #: Whether the user is playing Messenger game now
    in_game: Optional[bool] = None

    @classmethod
    def _from_orca_presence(cls, data)-> ActiveStatus:
        # TODO: Handle `c` and `vc` keys (Probably some binary data)
        return cls(active=data["p"] in [2, 3], last_active=data.get("l"), in_game=None)



from __future__ import annotations
from dataclasses import dataclass, field
from . import _plan
from ._thread import ThreadType, Thread
from typing import Any, Dict, Set, Optional


@dataclass(eq=False)
class Group(Thread):
    """Represents a Facebook group. Inherits `Thread`."""

    #: Set of the group thread's participant user IDs
    participants: Set[str] = field(default_factory=set)
    #: A dictionary, containing user nicknames mapped to their IDs
    nicknames: Optional[Dict[str, str]] = field(default_factory=dict)
    #: A :class:`ThreadColor`. The groups's message color
    color: Optional[str] = None
    #: The groups's default emoji
    emoji: Optional[str] = None
    #: Set containing user IDs of thread admins
    admins: Optional[Set[str]] = field(default_factory=set)
    #: True if users need approval to join
    approval_mode: Optional[bool] = None
    #: Set containing user IDs requesting to join
    approval_requests: Optional[Set[str]] = field(default_factory=set)
    #: Link for joining group
    join_link: Optional[str] = None

    def __post_init__(self):
        self.type = ThreadType.GROUP
        self.participants = set() if self.participants is None else self.participants
        self.nicknames = {} if self.nicknames is None else self.nicknames
        self.admins = set() if self.admins is None else self.admins
        self.approval_requests = set() if self.approval_requests is None else self.approval_requests



    @classmethod
    def _from_graphql(cls, data: Dict[str, Any])-> Group:
        if data.get("image") is None:
            data["image"] = {}
        c_info = cls._parse_customization_info(data)
        last_message_timestamp = None
        if "last_message" in data:
            last_message_timestamp = data["last_message"]["nodes"][0][
                "timestamp_precise"
            ]
        plan = None
        if data.get("event_reminders") and data["event_reminders"].get("nodes"):
            plan = _plan.Plan._from_graphql(data["event_reminders"]["nodes"][0])

        return cls(
            data["thread_key"]["thread_fbid"],
            participants=set(
                [
                    node["messaging_actor"]["id"]
                    for node in data["all_participants"]["nodes"]
                ]
            ),
            nicknames=c_info.get("nicknames"),
            color=c_info.get("color"),
            emoji=c_info.get("emoji"),
            admins=set([node.get("id") for node in data.get("thread_admins")]), #type: ignore
            approval_mode=bool(data.get("approval_mode"))
            if data.get("approval_mode") is not None
            else None,
            approval_requests=set(
                node["requester"]["id"]
                for node in data["group_approval_queue"]["nodes"]
            )
            if data.get("group_approval_queue")
            else None,
            join_link=data["joinable_mode"].get("link"),
            photo=data["image"].get("uri"),
            name=data["name"],
            message_count=data.get("messages_count"),
            last_message_timestamp=last_message_timestamp,
            plan=plan,
        )

    def _to_send_data(self)-> Dict[str, str]:
        return {"thread_fbid": self.uid}


@dataclass(eq=False)
class Room(Group):
    """Deprecated. Use `Group` instead."""
    #: True is room is not discoverable
    privacy_mode: Optional[bool] = None

    def __post_init__(self):
        self.type = ThreadType.ROOM



"""
For Type annotation you can import all models from here.
"""

__all__ = [
    "FBchatException", "FBchatFacebookError", "FBchatUserError", "FBchatPleaseRefresh",
    "FBchatInvalidParameters", "FBchatNotLoggedIn",
    "CustomEnum",

    "ThreadType", "ThreadLocation", "ThreadColor", "Thread", "Group", "Room",
    "TypingStatus", "User", "ActiveStatus", "Page",

    "EmojiSize", "MessageReaction", "Mention", "Message", "Sticker",
    "Attachment", "UnsentMessage", "ShareAttachment",
    "LocationAttachment", "LiveLocationAttachment", "FileAttachment", 
    "AudioAttachment", "ImageAttachment", "VideoAttachment",

    "QuickReply", "QuickReplyText", "QuickReplyLocation", "QuickReplyPhoneNumber", "QuickReplyEmail",

    "Poll", "PollOption", "GuestStatus", "Plan"

]
from ._core import CustomEnum
from ._exception import (
    FBchatException, 
    FBchatFacebookError, 
    FBchatUserError, 
    FBchatNotLoggedIn,
    FBchatInvalidParameters,
    FBchatPleaseRefresh
)
from ._thread import ThreadType, ThreadLocation, ThreadColor, Thread
from ._user import TypingStatus, User, ActiveStatus
from ._group import Group, Room
from ._page import Page
from ._message import EmojiSize, MessageReaction, Mention, Message
from ._attachment import Attachment, UnsentMessage, ShareAttachment
from ._sticker import Sticker
from ._location import LocationAttachment, LiveLocationAttachment
from ._file import FileAttachment, AudioAttachment, ImageAttachment, VideoAttachment
from ._quick_reply import (
    QuickReply,
    QuickReplyText,
    QuickReplyLocation,
    QuickReplyPhoneNumber,
    QuickReplyEmail,
)
from ._poll import Poll, PollOption
from ._plan import GuestStatus, Plan


__title__ = "fbchat-muqit"
__version__ = "1.0.2"
__description__ = "A powerful Facebook Messenger Chat API to interact with Facebook and Messenger. Easy to use fast and efficient library"

__copyright__ = "Copyright 2024 by Muhammad MuQiT"

__license__ = "GPL-V3.0"

__author__ = "Muhammad MuQiT"
__email__ = "togashiyuuta1111@gmail.com"



from __future__ import annotations
from dataclasses import dataclass
from typing import List, Optional


@dataclass(eq=False)
class Poll:
    """Represents a poll."""

    #: Title of the poll
    title: str 
    #: List of :class:`PollOption`, can be fetched with :func:`fbchat.Client.fetchPollOptions`
    options: List["PollOption"]
    #: Options count
    options_count: Optional[int] = None
    #: ID of the poll
    uid: Optional[int] = None

    @classmethod
    def _from_graphql(cls, data)-> Poll:
        return cls(
            uid=int(data["id"]),
            title=data.get("title") if data.get("title") else data.get("text"),
            options=[PollOption._from_graphql(m) for m in data.get("options")],
            options_count=data.get("total_count"),
        )


@dataclass(eq=False)
class PollOption:
    """Represents a poll option."""

    #: Text of the poll option
    text: str
    #: Whether vote when creating or client voted
    vote: Optional[bool] = False
    #: ID of the users who voted for this poll option
    voters: Optional[List] = None
    #: Votes count
    votes_count: Optional[int] = None
    #: ID of the poll option
    uid: Optional[int] = None

    @classmethod
    def _from_graphql(cls, data)-> PollOption:
        if data.get("viewer_has_voted") is None:
            vote = None
        elif isinstance(data["viewer_has_voted"], bool):
            vote = data["viewer_has_voted"]
        else:
            vote: Optional[bool] = data["viewer_has_voted"] == "true"
        return cls(
            uid=int(data["id"]),
            text=data.get("text"),
            vote=vote,
            voters=(
                [m.get("node").get("id") for m in data.get("voters").get("edges")]
                if isinstance(data.get("voters"), dict)
                else data.get("voters")
            ),
            votes_count=(
                data.get("voters").get("count")
                if isinstance(data.get("voters"), dict)
                else data.get("total_count")
            ),
        )



import json
from time import time
from random import random
from contextlib import contextmanager
from mimetypes import guess_type
from os.path import basename
from urllib.parse import parse_qs, urlparse
from .models import (
FBchatException,
    FBchatFacebookError,
    FBchatInvalidParameters,
    FBchatNotLoggedIn,
    FBchatPleaseRefresh,
)

def now():
    return int(time() * 1000)


def json_minimal(data):
    """Get JSON data in minimal form."""
    return json.dumps(data, separators=(",", ":"))


def strip_json_cruft(text):
    """Removes `for(;;);` (and other cruft) that preceeds JSON responses."""
    try:
        return text[text.index("{") :]
    except ValueError:
        raise FBchatException("No JSON object found: {!r}".format(text))


def get_decoded_r(r):
    return get_decoded(r._content)


def get_decoded(content):
    return content.decode("utf-8")


def parse_json(content):
    try:
        return json.loads(content)
    except ValueError:
        raise FBchatFacebookError("Error while parsing JSON: {!r}".format(content))


def digitToChar(digit):
    if digit < 10:
        return str(digit)
    return chr(ord("a") + digit - 10)


def str_base(number, base):
    if number < 0:
        return "-" + str_base(-number, base)
    (d, m) = divmod(number, base)
    if d > 0:
        return str_base(d, base) + digitToChar(m)
    return digitToChar(m)


def generateMessageID(client_id=None):
    k = now()
    l = int(random() * 4294967295)
    return "<{}:{}-{}@mail.projektitan.com>".format(k, l, client_id)


def getSignatureID():
    return hex(int(random() * 2147483648))


def generateOfflineThreadingID():
    ret = now()
    value = int(random() * 4294967295)
    string = ("0000000000000000000000" + format(value, "b"))[-22:]
    msgs = format(ret, "b") + string
    return str(int(msgs, 2))


def handle_payload_error(j):
    if "error" not in j:
        return
    error = j["error"]
    if j["error"] == 1357001:
        error_cls = FBchatNotLoggedIn
    elif j["error"] == 1357004:
        error_cls = FBchatPleaseRefresh
    elif j["error"] in (1357031, 1545010, 1545003):
        error_cls = FBchatInvalidParameters
    else:
        error_cls = FBchatFacebookError
    # TODO: Use j["errorSummary"]
    # "errorDescription" is in the users own language!
    raise error_cls(
        "Error #{} when sending request: {}".format(error, j["errorDescription"]),
        fb_error_code=error,
        fb_error_message=j["errorDescription"],
    )


def handle_graphql_errors(j):
    errors = []
    if j.get("error"):
        errors = [j["error"]]
    if "errors" in j:
        errors = j["errors"]
    if errors:
        error = errors[0]  # TODO: Handle multiple errors
        # TODO: Use `summary`, `severity` and `description`
        raise FBchatFacebookError(
            "GraphQL error #{}: {} / {!r}".format(
                error.get("code"), error.get("message"), error.get("debug_info")
            ),
            fb_error_code=error.get("code"),
            fb_error_message=error.get("message"),
        )


def check_request(r):
    check_http_code(r.status_code)
    content = get_decoded_r(r)
    check_content(content)
    return content

async def async_check_request(r):
    check_http_code(r.status)
    content = r.text(encoding="utf-8")
    check_content(content)
    return content


def check_http_code(code):
    msg = "Error when sending request: Got {} response.".format(code)
    if code == 404:
        raise FBchatFacebookError(
            msg + " This is either because you specified an invalid URL, or because"
            " you provided an invalid id (Facebook usually requires integer ids).",
            request_status_code=code,
        )
    if 400 <= code < 600:
        raise FBchatFacebookError(msg, request_status_code=code)


def check_content(content, as_json=True):
    if content is None or len(content) == 0:
        raise FBchatFacebookError("Error when sending request: Got empty response")


def to_json(content):
    content = strip_json_cruft(content)
    j = parse_json(content)
    return j


def get_jsmods_require(j, index):
    if j.get("jsmods") and j["jsmods"].get("require"):
        try:
            return j["jsmods"]["require"][0][index][0]
        except (KeyError, IndexError):
            print(
                "Error when getting jsmods_require: "
                "{}. Facebook might have changed protocol".format(j)
            )
    return None


def require_list(list_):
    if isinstance(list_, list):
        return set(list_)
    else:
        return set([list_])


def mimetype_to_key(mimetype):
    if not mimetype:
        return "file_id"
    if mimetype == "image/gif":
        return "gif_id"
    x = mimetype.split("/")
    if x[0] in ["video", "image", "audio"]:
        return "%s_id" % x[0]
    return "file_id"




@contextmanager
def get_files_from_paths(filenames):
    files = []
    for filename in filenames:
        files.append(
            (basename(filename), open(filename, "rb"), guess_type(filename)[0])
        )
    yield files
    for fn, fp, ft in files:
        fp.close()


def get_url_parameters(url, *args):
    params = parse_qs(urlparse(url).query)
    return [params[arg][0] for arg in args if params.get(arg)]


def get_url_parameter(url, param):
    return get_url_parameters(url, param)[0]


def prefix_url(url):
    if url.startswith("/"):
        return "https://www.facebook.com" + url
    return url



import logging
from aenum import Enum, extend_enum

log = logging.getLogger("client")


class CustomEnum(Enum): #type: ignore
    """Used internally by ``fbchat`` to support enumerations"""

    def __repr__(self):
        # For documentation:
        return f"{type(self).__name__}.{self.name}"

    @classmethod
    def _extend_if_invalid(cls, value):
        try:
            return cls(value)
        except ValueError:
            log.warning(
                "Failed parsing {.__name__}({!r}). Extending enum.".format(cls, value)
            )
            extend_enum(cls, f"UNKNOWN_{value}".upper(), value)
            return cls(value)



from .models import (
    ThreadType, ThreadLocation, ThreadColor, Thread,
    TypingStatus, User, ActiveStatus, 
    Group, Room, Page, CustomEnum,
    EmojiSize, Sticker, MessageReaction, Mention, Message, UnsentMessage,
    Attachment, ShareAttachment, LocationAttachment, LiveLocationAttachment,
    FileAttachment, AudioAttachment, ImageAttachment, VideoAttachment,
    QuickReply, QuickReplyText, QuickReplyLocation, QuickReplyPhoneNumber,
    QuickReplyEmail,
    Poll, PollOption,
    GuestStatus, Plan,
    FBchatException, FBchatFacebookError, FBchatUserError
)
from .n_client import Client
from .n_state import State

__title__ = "fbchat-muqit"
__version__ = "1.0.3"
__description__ = "A powerful Facebook Messenger Chat API to interact with Facebook and Messenger. Easy to use fast and efficient library"

__copyright__ = "Copyright 2024 by Muhammad MuQiT"

__license__ = "GPL-V3.0"

__author__ = "Muhammad MuQiT"
__email__ = "togashiyuuta1111@gmail.com"

__all__ = [
    "Client", "State",

    "FBchatException", "FBchatFacebookError", "FBchatUserError", "CustomEnum",

    "ThreadType", "ThreadLocation", "ThreadColor", "Thread", "Group", "Room",
    "TypingStatus", "User", "ActiveStatus", "Page",

    "EmojiSize", "MessageReaction", "Mention", "Message", "Sticker",
    "Attachment", "UnsentMessage", "ShareAttachment",
    "LocationAttachment", "LiveLocationAttachment", "FileAttachment", 
    "AudioAttachment", "ImageAttachment", "VideoAttachment",

    "QuickReply", "QuickReplyText", "QuickReplyLocation", "QuickReplyPhoneNumber", "QuickReplyEmail",

    "Poll", "PollOption", "GuestStatus", "Plan" 

]
